#!/usr/bin/perl

use strict;
use warnings;

use Data::Dumper;
use LWP::UserAgent;
use HTTP::Request;
use Xorux_lib;
use JSON qw(encode_json decode_json);
use Time::Local;
use XML::Simple;

$XML::Simple::PREFERRED_PARSER = 'XML::Parser';

my $storage_name   = $ENV{'STORAGE_NAME'}         ||= "";
my $vplex_ssh_ip   = $ENV{'EMC_VPLEX_SSH_IP'}     ||= "";
my $vplex_ssh_port = $ENV{'EMC_VPLEX_SSH_PORT'}   ||= "";
my $vplex_ssh_user = $ENV{'EMC_VPLEX_SSH_USER'}   ||= "";
my $vplex_api      = $ENV{'EMC_VPLEX_API'}        ||= "";
my $vplex_api_ip   = $ENV{'EMC_VPLEX_API_IP'}     ||= "";
my $vplex_api_port = $ENV{'EMC_VPLEX_API_PORT'}   ||= "";
my $cluster_id     = $ENV{'EMC_VPLEX_CLUSTER_ID'} ||= "";
my $user           = $ENV{'EMC_VPLEX_USER'}       ||= "";
my $pass           = $ENV{'EMC_VPLEX_PASSWD'}     ||= "";
my $SSH            = $ENV{'SSH'}                  ||= "";
my $uname          = $ENV{'UNAME_FOR_VPLEX'}      ||= "";
my $sample_rate    = $ENV{'SAMPLE_RATE'}          ||= 300;

$pass = Xorux_lib::unobscure_password($pass);

if ( !defined $storage_name || $storage_name eq '' ) {
  error( "Storage NAME must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $vplex_ssh_ip || $vplex_ssh_ip eq '' ) {
  error( "VPLEX mgmt server IP or hostname must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $cluster_id || $cluster_id eq '' ) {
  error( "Cluster ID must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $vplex_ssh_port || $vplex_ssh_port eq '' ) {
  error( "SSH port to mgmt server must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $vplex_ssh_user || $vplex_ssh_user eq '' ) {
  error( "Mgmt server user must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}
if ( !defined $SSH || $SSH eq '' ) {
  error( "SSH string must be set at stor2rrd.cfg! $!" . __FILE__ . ":" . __LINE__ ) && exit;
}

# use API
my $url = "";
if ( defined $vplex_api && $vplex_api eq "1" ) {
  if ( !defined $vplex_api_ip || $vplex_api_ip eq '' ) {
    error( "API IP must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }
  if ( !defined $vplex_api_port || $vplex_api_port eq '' ) {
    error( "API port must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }
  if ( !defined $user || $user eq '' ) {
    error( "API USER must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }
  if ( !defined $pass || $pass eq '' ) {
    error( "API PASSWORD must be set! $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  $url = "https://$vplex_api_ip:$vplex_api_port/vplex";
} ## end if ( defined $vplex_api...)

my $basedir      = $ENV{INPUTDIR};
my $webdir       = $ENV{WEBDIR};
my $wrkdir       = "$basedir/data";
my $tmpdir       = "$basedir/tmp";
my $monitors_dir = "$wrkdir/$storage_name/iostats";

if ( !-d "$monitors_dir" ) {
  mkdir( "$monitors_dir", 0755 ) || error( "Cannot mkdir $monitors_dir: $!" . __FILE__ . ":" . __LINE__ ) && exit;
}

my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime();
my $time_to_output_file = sprintf( "%4d%02d%02d_%02d%02d", $year + 1900, $month + 1, $day, $hour, $min );

my $perf_out = "$wrkdir/$storage_name/$storage_name\_vplexperf_$time_to_output_file.out.tmp";
my $conf_out = "$wrkdir/$storage_name/$storage_name\_vplexconf_$time_to_output_file.out.tmp";

my %parameters;
my %keys;
my %data;
my %monitors;

download_monitors();
parse_monitors();
print_perf();


# Health status
# API required
my $health_status_html      = "$webdir/$storage_name/health_status.html";
my $health_status_time_diff = file_time_diff($health_status_html);
if ( defined $vplex_api && $vplex_api eq "1" && ( !-f $health_status_html || $health_status_time_diff > 7199 ) ) {
  health_status();
}


# Array capacity
# API required
# run this cmd only once per day
my $capacity_file      = "$monitors_dir/capacity-clusters.txt";
my $cap_file_time_diff = file_time_diff($capacity_file);
if ( defined $vplex_api && $vplex_api eq "1" && ( !-f $capacity_file || $cap_file_time_diff > 86399 ) ) {
  capacity_array();
}


#
# print CONF file
#
my $config_html   = "$wrkdir/$storage_name/config.html";
my $cfg_time_diff = file_time_diff($config_html);

# use API
if ( defined $vplex_api && $vplex_api eq "1" ) {
  if ( !-f $config_html || $cfg_time_diff > 86399 ) {
    check_clusters();

    if ( defined $data{'conf'}{'cluster'}{$cluster_id} ) {
      #
      # check/download configdump.xml
      #
      my $configdump_filename = "stor2rrd_$cluster_id\_configdump_$uname.xml";
      my $configdump_file     = "$monitors_dir/$configdump_filename";

      print_info("CONF : download_configdump : $configdump_filename");
      my $configdump_state = download_configdump("$configdump_filename");    # 0/1 -> not exists/exist

      if ( $configdump_state eq "1" ) {
        print_info("CONF : download_configdump : $configdump_file was copied successfully");

        parse_configdump("$configdump_file");

        print_conf_from_api();
        print_info("CONF : $conf_out was created successfully");
      } ## end if ( $configdump_state...)
    } ## end if ( defined $data{'conf'...})
    else {
      error( "Configured Cluster name \"$cluster_id\" was not found via rest api! : $!" . __FILE__ . ":" . __LINE__ );

      my $avail_clusters = "";
      if ( defined $data{'conf'}{'cluster'} && defined $data{'conf'}{'cluster'} ne "" ) {
        my $clust_idx = 0;
        foreach my $avail_cluster ( sort keys( %{ $data{'conf'}{'cluster'} } ) ) {
          $clust_idx++;
          if ( $clust_idx == 1 ) {
            $avail_clusters = "\"$avail_cluster\"";
          }
          else {
            $avail_clusters .= " , \"$avail_cluster\"";
          }
        } ## end foreach my $avail_cluster (...)
      } ## end if ( defined $data{'conf'...})

      if ( defined $avail_clusters && $avail_clusters ne "" ) {
        error( "Availaible Clusters via rest api: $avail_clusters : $!" . __FILE__ . ":" . __LINE__ );
      }
    } ## end else [ if ( defined $data{'conf'...})]
  } ## end if ( !-f $config_html ...)
} ## end if ( defined $vplex_api...)
else {
  if ( !-f $config_html || $cfg_time_diff > 3599 ) {
    print_conf_from_perf();
    print_info("CONF : $conf_out was created successfully");
  }
} ## end else [ if ( defined $vplex_api...)]



#print Dumper \%monitors;
#print Dumper \%data;
#print Dumper (\%{$data{'perf'}{'virtual volumes'}});
exit;

sub parse_configdump {
  my $xml_file = shift;
  my $xml      = "";

  eval { $xml = XMLin( "$xml_file", ForceArray => 1 ); };
  if ($@) {
    error("XML parsing error $xml_file: $@");
    error("Try to parse only required sections: cluster, views, virtual-volumes");
    $xml = parse_xml_sections($xml_file);
  }

  # sections:
  # ---------
  # extents
  # cluster
  # system-volumes
  # initiators
  # virtual-volumes
  # version
  # time
  # views
  # ports
  # disks
  # top-level-devices

  #print Dumper $xml->{'disks'};

  # parse cluster info
  if ( defined $xml->{'cluster'} && $xml->{'cluster'} ne "" && ref( $xml->{'cluster'} ) eq "ARRAY" ) {
    foreach ( @{ $xml->{'cluster'} } ) {
      if ( defined $_ && $_ ne "" && ref($_) eq "HASH" ) {
        if ( defined $_->{'name'} && $_->{'name'} eq $cluster_id ) {
          if ( defined $_->{'managementIP'} && $_->{'managementIP'} ne "" ) {
            $data{'conf'}{'cluster'}{ $_->{'name'} }{'managementIP'} = $_->{'managementIP'};
          }
          if ( defined $_->{'version'} && $_->{'version'} ne "" ) {
            $data{'conf'}{'cluster'}{ $_->{'name'} }{'version'} = $_->{'version'};
          }
          if ( defined $_->{'top-level-assembly'} && $_->{'top-level-assembly'} ne "" ) {
            $data{'conf'}{'cluster'}{ $_->{'name'} }{'top-level-assembly'} = $_->{'top-level-assembly'};
          }
        } ## end if ( defined $_->{'name'...})
      } ## end if ( defined $_ && $_ ...)
    } ## end foreach ( @{ $xml->{'cluster'...}})
  } ## end if ( defined $xml->{'cluster'...})

  # parse storage views info (Hosts -> Volumes)
  if ( defined $xml->{'views'} && $xml->{'views'} ne "" && ref( $xml->{'views'} ) eq "ARRAY" ) {
    foreach ( @{ $xml->{'views'} } ) {
      if ( defined $_->{'view'} && $_->{'view'} ne "" && ref( $_->{'view'} ) eq "HASH" ) {
        foreach my $storage_view ( sort keys( %{ $_->{'view'} } ) ) {

          #print "$storage_view\n";
          $data{'conf'}{'storage_view'}{$storage_view}{'name'} = $storage_view;

          if ( defined $_->{'view'}{$storage_view}{'viewVolumes'} && $_->{'view'}{$storage_view}{'viewVolumes'} ne "" && ref( $_->{'view'}{$storage_view}{'viewVolumes'} ) eq "ARRAY" ) {
            foreach ( @{ $_->{'view'}{$storage_view}{'viewVolumes'} } ) {
              if ( defined $_ && $_ ne "" && ref($_) eq "HASH" ) {
                if ( defined $_->{'viewVolume'} && $_->{'viewVolume'} ne "" ) {
                  foreach my $volume ( sort keys( %{ $_->{'viewVolume'} } ) ) {

                    #print "   $volume\n";
                    $data{'conf'}{'storage_view'}{$storage_view}{'virtual-volumes'}{$volume} = $volume;
                  }
                } ## end if ( defined $_->{'viewVolume'...})
              } ## end if ( defined $_ && $_ ...)
            } ## end foreach ( @{ $_->{'view'}{$storage_view...}})
          } ## end if ( defined $_->{'view'...})
        } ## end foreach my $storage_view ( ...)
      } ## end if ( defined $_->{'view'...})
    } ## end foreach ( @{ $xml->{'views'...}})
  } ## end if ( defined $xml->{'views'...})

  # parse virtual-volumes info
  if ( defined $xml->{'virtual-volumes'} && $xml->{'virtual-volumes'} ne "" && ref( $xml->{'virtual-volumes'} ) eq "ARRAY" ) {
    foreach ( @{ $xml->{'virtual-volumes'} } ) {
      if ( defined $_->{'volume'} && $_->{'volume'} ne "" && ref( $_->{'volume'} ) eq "HASH" ) {
        foreach my $volume ( sort keys( %{ $_->{'volume'} } ) ) {

          #print "$volume\n";
          $data{'conf'}{'virtual-volumes'}{$volume}{'name'} = $volume;
          if ( defined $_->{'volume'}{$volume}{'size'} && isdigit( $_->{'volume'}{$volume}{'size'} ) ) {
            $data{'conf'}{'virtual-volumes'}{$volume}{'size'} = $_->{'volume'}{$volume}{'size'};
          }
          if ( defined $_->{'volume'}{$volume}{'id'} && $_->{'volume'}{$volume}{'id'} ne "" ) {
            $data{'conf'}{'virtual-volumes'}{$volume}{'id'} = $_->{'volume'}{$volume}{'id'};
          }
          if ( defined $_->{'volume'}{$volume}{'vpdIdentifier'} && $_->{'volume'}{$volume}{'vpdIdentifier'} ne "" ) {
            $data{'conf'}{'virtual-volumes'}{$volume}{'vpdIdentifier'} = $_->{'volume'}{$volume}{'vpdIdentifier'};
          }
          if ( defined $_->{'volume'}{$volume}{'locality'} && $_->{'volume'}{$volume}{'locality'} ne "" ) {
            $data{'conf'}{'virtual-volumes'}{$volume}{'locality'} = $_->{'volume'}{$volume}{'locality'};
          }
          if ( defined $_->{'volume'}{$volume}{'supportingDevice'} && $_->{'volume'}{$volume}{'supportingDevice'} ne "" ) {
            $data{'conf'}{'virtual-volumes'}{$volume}{'supportingDevice'} = $_->{'volume'}{$volume}{'supportingDevice'};
          }
          if ( defined $_->{'volume'}{$volume}{'cacheMode'} && $_->{'volume'}{$volume}{'cacheMode'} ne "" ) {
            $data{'conf'}{'virtual-volumes'}{$volume}{'cacheMode'} = $_->{'volume'}{$volume}{'cacheMode'};
          }
        } ## end foreach my $volume ( sort keys...)
      } ## end if ( defined $_->{'volume'...})
    } ## end foreach ( @{ $xml->{'virtual-volumes'...}})
  } ## end if ( defined $xml->{'virtual-volumes'...})

  return 1;
} ## end sub parse_configdump

sub parse_xml_sections {
  my $xml_file = shift;
  my %xml_out;

  my @cluster_data;
  my @volumes_data;
  my @views_data;

  my $cluster_section = 0;
  my $volumes_section = 0;
  my $views_section   = 0;

  open( XML, "< $xml_file" ) || error( "Couldn't open file $xml_file $!" . __FILE__ . ":" . __LINE__ ) && return '';

  while ( my $line = <XML> ) {
    chomp $line;

    # virtual-volumes info
    if ( $line =~ m/^<virtual-volumes>/ ) { $volumes_section++; }
    if ( $volumes_section == 1 ) {
      push(@volumes_data, "$line\n");
    }
    if ( $line =~ m/^<\/virtual-volumes>/ ) { $volumes_section++; }

    # cluster info
    if ( $line =~ m/^<cluster/ ) { $cluster_section++; }
    if ( $cluster_section == 1 ) {
      push(@cluster_data, "$line\n");
    }
    if ( $line =~ m/^<\/cluster>/ ) { $cluster_section++; }

    # views info
    if ( $line =~ m/^<views>/ ) { $views_section++; }
    if ( $views_section == 1 ) {
      push(@views_data, "$line\n");
    }
    if ( $line =~ m/^<\/views>/ ) { $views_section++; }
  }
  close(XML);


  # volumes
  if ( $volumes_section == 2 ) {
    eval { @{$xml_out{'virtual-volumes'}} = XMLin( "@volumes_data", ForceArray => 1 );  };
    if ($@) {
      error("XML parsing error section \"virtual-volumes\": $@");
      print_info("CONF : create conf file from perf stats! Downloaded configdump file is corrupted!");
      print_conf_from_perf();
      print_info("CONF : $conf_out was created successfully");
      exit;
    }
  }
  else {
    error("XML parsing error section \"virtual-volumes\" : section START or END does not found!  $@");
    print_info("CONF : create conf file from perf stats! Downloaded configdump file is corrupted!");
    print_conf_from_perf();
    print_info("CONF : $conf_out was created successfully");
    exit;
  }

  # cluster
  if ( $cluster_section == 2 ) {
    eval { @{$xml_out{'cluster'}} = XMLin( "@cluster_data", ForceArray => 1 );  };
    if ($@) {
      error("XML parsing error section \"cluster\": $@");
    }
  }
  else {
    error("XML parsing error section \"cluster\" : section START or END does not found!  $@");
  }

  # views
  if ( $views_section == 2 ) {
    my $xml_views = "";
    eval { @{$xml_out{'views'}} = XMLin( "@views_data", ForceArray => 1 );  };
    if ($@) {
      error("XML parsing error section \"views\": $@");
    }
  }
  else {
    error("XML parsing error section \"views\" : section START or END does not found!  $@");
  }

  return \%xml_out;
}

sub print_conf_from_perf {
  open( CONF, ">$conf_out" ) || error( "Couldn't open file $conf_out $!" . __FILE__ . ":" . __LINE__ ) && exit;

  # Virtual volumes
  print CONF "Volume Level Configuration\n";
  print CONF "---------------------\n";
  print CONF "Volume ID,Virtual Volume,Director,size,vpdIdentifier,locality,supportingDevice,cacheMode\n";

  my $vvol_idx = 0;
  foreach my $vvol ( sort keys( %{ $data{'perf'}{'virtual volumes'} } ) ) {
    $vvol_idx++;

    my $directors = "";

    if ( defined $data{'perf'}{'virtual volumes'}{$vvol} && $data{'perf'}{'virtual volumes'}{$vvol} ne '' ) {
      my $dir_idx = 0;
      foreach my $dir ( sort keys( %{ $data{'perf'}{'virtual volumes'}{$vvol} } ) ) {

        #print "     $vvol\n";
        $dir_idx++;
        if ( defined $directors && $directors ne '' ) {
          $directors .= " $dir";
        }
        else {
          $directors = "$dir";
        }
      } ## end foreach my $dir ( sort keys...)
    } ## end if ( defined $data{'perf'...})
    print CONF "$vvol,$vvol,$directors,,,,,\n";
  } ## end foreach my $vvol ( sort keys...)
  print CONF "\n";
  close(CONF);

  return 1;
} ## end sub print_conf_from_perf

sub print_conf_from_api {
  open( CONF, ">$conf_out" ) || error( "Couldn't open file $conf_out $!" . __FILE__ . ":" . __LINE__ ) && exit;

  # cluster info
  my $version = "";
  my $mgmt_ip = "";
  my $serial  = "";
  if ( defined $data{'conf'}{'cluster'}{$cluster_id}{'managementIP'} && $data{'conf'}{'cluster'}{$cluster_id}{'managementIP'} ne "" ) {
    $mgmt_ip = $data{'conf'}{'cluster'}{$cluster_id}{'managementIP'};
  }
  if ( defined $data{'conf'}{'cluster'}{$cluster_id}{'version'} && $data{'conf'}{'cluster'}{$cluster_id}{'version'} ne "" ) {
    $version = $data{'conf'}{'cluster'}{$cluster_id}{'version'};
  }
  if ( defined $data{'conf'}{'cluster'}{$cluster_id}{'top-level-assembly'} && $data{'conf'}{'cluster'}{$cluster_id}{'top-level-assembly'} ne "" ) {
    $serial = $data{'conf'}{'cluster'}{$cluster_id}{'top-level-assembly'};
  }

  print CONF "Configuration Data\n";
  print CONF "------------------\n";
  print CONF "managementIP : $mgmt_ip\n";
  print CONF "version      : $version\n";
  print CONF "serial       : $serial\n";
  print CONF "\n";


  # Virtual volumes
  print CONF "Volume Level Configuration\n";
  print CONF "---------------------\n";
  print CONF "Volume ID,Virtual Volume,Director,size,vpdIdentifier,locality,supportingDevice,cacheMode\n";

  my $vvol_idx = 0;
  my $cap_tot  = "";
  foreach my $vvol ( sort keys( %{ $data{'conf'}{'virtual-volumes'} } ) ) {

    #print "   $dir\n";
    $vvol_idx++;

    my $directors        = "";
    my $size             = "";
    my $vpdIdentifier    = "";
    my $locality         = "";
    my $supportingDevice = "";
    my $cacheMode        = "";

    if ( defined $data{'conf'}{'virtual-volumes'}{$vvol}{'size'} && $data{'conf'}{'virtual-volumes'}{$vvol}{'size'} ne "" ) {
      $size = $data{'conf'}{'virtual-volumes'}{$vvol}{'size'};
      if ( isdigit($size) && defined $cap_tot && $cap_tot eq "" ) {
        $cap_tot = $size;
      }
      if ( isdigit($size) && defined $cap_tot && isdigit($cap_tot) ) {
        $cap_tot = $cap_tot + $size;
      }
    }
    if ( defined $data{'conf'}{'virtual-volumes'}{$vvol}{'vpdIdentifier'} && $data{'conf'}{'virtual-volumes'}{$vvol}{'vpdIdentifier'} ne "" ) {
      $vpdIdentifier = $data{'conf'}{'virtual-volumes'}{$vvol}{'vpdIdentifier'};
    }
    if ( defined $data{'conf'}{'virtual-volumes'}{$vvol}{'locality'} && $data{'conf'}{'virtual-volumes'}{$vvol}{'locality'} ne "" ) {
      $locality = $data{'conf'}{'virtual-volumes'}{$vvol}{'locality'};
    }
    if ( defined $data{'conf'}{'virtual-volumes'}{$vvol}{'supportingDevice'} && $data{'conf'}{'virtual-volumes'}{$vvol}{'supportingDevice'} ne "" ) {
      $supportingDevice = $data{'conf'}{'virtual-volumes'}{$vvol}{'supportingDevice'};
    }
    if ( defined $data{'conf'}{'virtual-volumes'}{$vvol}{'cacheMode'} && $data{'conf'}{'virtual-volumes'}{$vvol}{'cacheMode'} ne "" ) {
      $cacheMode = $data{'conf'}{'virtual-volumes'}{$vvol}{'cacheMode'};
    }

    if ( defined $data{'perf'}{'virtual volumes'}{$vvol} && $data{'perf'}{'virtual volumes'}{$vvol} ne '' ) {
      my $dir_idx = 0;
      foreach my $dir ( sort keys( %{ $data{'perf'}{'virtual volumes'}{$vvol} } ) ) {

        #print "     $vvol\n";
        $dir_idx++;
        if ( defined $directors && $directors ne '' ) {
          $directors .= " $dir";
        }
        else {
          $directors = "$dir";
        }
      } ## end foreach my $dir ( sort keys...)
    } ## end if ( defined $data{'perf'...})
    print CONF "$vvol,$vvol,$directors,$size,$vpdIdentifier,$locality,$supportingDevice,$cacheMode\n";
  } ## end foreach my $vvol ( sort keys...)
  print CONF "\n";


  # Storage view (HOST)
  print CONF "Storage view Level Configuration\n";
  print CONF "---------------------\n";
  print CONF "Host name,Storage view,virtual-volumes\n";

  my $st_view_idx = 0;
  foreach my $storage_view ( sort keys( %{ $data{'conf'}{'storage_view'} } ) ) {

    #print "   $dir\n";
    $st_view_idx++;

    my $vvols = "";

    if ( defined $data{'conf'}{'storage_view'}{$storage_view}{'virtual-volumes'} && $data{'conf'}{'storage_view'}{$storage_view}{'virtual-volumes'} ne "" ) {
      my $vvol_idx = 0;
      foreach my $vvol ( sort keys( %{ $data{'conf'}{'storage_view'}{$storage_view}{'virtual-volumes'} } ) ) {
        $vvol_idx++;
        if ( defined $vvols && $vvols ne "" ) {
          $vvols .= " $vvol";
        }
        else {
          $vvols = "$vvol";
        }
      } ## end foreach my $vvol ( sort keys...)
    } ## end if ( defined $data{'conf'...})
    print CONF "$storage_view,$storage_view,$vvols\n";
  } ## end foreach my $storage_view ( ...)
  print CONF "\n";


  # Array/Pool capacity
  if ( -f $capacity_file ) {
    print CONF "Array Level Configuration\n";
    print CONF "---------------------\n";
    print CONF "Timestamp,cap total (GiB),cap used (GiB),cap free (GiB)\n";

    open( CAP, "<$capacity_file" ) || error( "Couldn't open file $capacity_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
    my @lines = <CAP>;
    close(CAP);

    # Unclaimed storage_volumes : 2
    # Claimed disk capacity(GiB) : 2000.01
    # Unclaimed disk capacity (GiB) : 138.34
    # Exported volumes : 1718
    # Claimed storage_volumes : 1
    # Unexported volumes : 58
    # Exported volume capacity (GiB) : 581359.38
    # Used storage-volume capacity (GiB) : 594932.25
    # Unexported volume capacity (GiB) : 13432.89
    # Used storage_volumes : 1779
    # Cluster : VPLXAT-0624

    my $unclaim_disk = "";
    my $claim_disk   = "";
    my $used_st_vol  = "";

    foreach my $line (@lines) {
      chomp $line;

      if ( $line =~ m/^Unclaimed disk capacity \(GiB\)=/ ) {
        $line =~ s/^Unclaimed disk capacity \(GiB\)=//;
        if ( isdigit($line) ) {
          $unclaim_disk = $line;
        }
      }
      if ( $line =~ m/^Claimed disk capacity\(GiB\)=/ ) {
        $line =~ s/^Claimed disk capacity\(GiB\)=//;
        if ( isdigit($line) ) {
          $claim_disk = $line;
        }
      }
      if ( $line =~ m/^Used storage-volume capacity \(GiB\)=/ ) {
        $line =~ s/^Used storage-volume capacity \(GiB\)=//;
        if ( isdigit($line) ) {
          $used_st_vol = $line;
        }
      }
    }

    # Total = Unclaimed disk c. + Claimed disk c. + Used storage-volume c
    # Used  = Used storage-volume capacity (GiB)
    # Free  = Capacity (total) - Used

    my $cap_tot  = "";
    my $cap_used = "";
    my $cap_free = "";

    if ( defined $unclaim_disk && isdigit($unclaim_disk) && defined $claim_disk && isdigit($claim_disk) && defined $used_st_vol && isdigit($used_st_vol) ) {
      $cap_tot  = $unclaim_disk + $claim_disk + $used_st_vol;
      $cap_used = $used_st_vol;
      $cap_free = $cap_tot - $cap_used;
    }
    my $act_time = time();

    print CONF "$act_time,$cap_tot,$cap_used,$cap_free\n";
    print CONF "\n";

    print CONF "Pool Level Configuration\n";
    print CONF "---------------------\n";
    print CONF "Pool Name,Pool ID,cap total (GiB),cap used (GiB),cap free (GiB)\n";
    print CONF "Array Total,array_total,$cap_tot,$cap_used,$cap_free\n";
    print CONF "\n";
  }

  close(CONF);

  return 1;
} ## end sub print_conf_from_api

sub capacity_array {
  my $cmd_status_file = "$monitors_dir/capacity-clusters_commandresult";

  if ( -f $cmd_status_file ) {
    # check last update
    my $time_diff = file_time_diff($cmd_status_file);
    if ( $time_diff > 7200 ) {
      unlink("$cmd_status_file");
      error( "Something wrong with \"report capacity-clusters\" cmd! This file is older than 2 hours, this should not happen! \"$cmd_status_file\" was removed...  $!" . __FILE__ . ":" . __LINE__ ) && return;
    }

    open( STAT, "<$cmd_status_file" ) || error( "Couldn't open file $cmd_status_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
    my $url4cmdstatus = <STAT>;
    chomp $url4cmdstatus;
    close(STAT);

    #check URL
    if ( $url4cmdstatus !~ m/\/vplex\/commandresult\// ) {
      error( "Array capacity : something wrong with this URL: \"$url4cmdstatus\" $!" . __FILE__ . ":" . __LINE__ );
      error( "Array capacity : \"$cmd_status_file\" file will be removed! $!" . __FILE__ . ":" . __LINE__ );
      unlink("$cmd_status_file");
    }

    #check status
    my $result = new_get_request_configdump("$url4cmdstatus");
    if ( defined $result && $result ne "0" ) {
      if ( $result =~ m/Command execution continues to be under progress/ ) {    # still working..
        print_info("Array capacity : Command execution continues to be under progress");
      }
      else {
        if ( defined $result->{'response'}{'custom-data'} && $result->{'response'}{'custom-data'} ne '' ) {    # file is finished
          print_info("Array capacity : Command was completed successfully");
          #print Dumper $result;
          unlink("$cmd_status_file");

          # parse and print cmd output
          my @out = split("\n", $result->{'response'}{'custom-data'});
          print_capacity_array(\@out);
        }
      }
    }
  }
  else {
    print_info("Array capacity : run cmd via API...");
    my $json_b = "{\"args\":\"\"}";
    my $result = new_post_request_configdump( "$url/report+capacity-clusters", $json_b );

    #print Dumper $result;
    if ( defined $result && $result ne "0" ) {
      if ( $result =~ m/Command execution taking longer than 60 seconds/ ) {

        #print "$result\n";
        my ( undef, $cmd_status_url ) = split( ",", $result );
        chomp $cmd_status_url;

        open( STAT, ">$cmd_status_file" ) || error( "Couldn't open file $cmd_status_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
        print STAT "$cmd_status_url\n";
        close(STAT);
      } ## end if ( $result =~ m/Command execution taking longer than 60 seconds/)
      else {
        if ( defined $result->{'response'}{'custom-data'} && $result->{'response'}{'custom-data'} ne '' ) {    # file is finished
          print_info("Array capacity : Command was completed successfully");
          #print Dumper $result;
          unlink("$cmd_status_file");

          # parse and print cmd output
          my @out = split("\n", $result->{'response'}{'custom-data'});
          print_capacity_array(\@out);
        }
      }
    }
  } ## end else [ if ( -f $cfgdump_cmd_status_file)]

  return 1;
}

sub print_capacity_array {
  my $cmd_out = shift;
  my @data    = @{$cmd_out};

  #$capacity_file
  my %cap_data;
  my @header;
  foreach my $line (@data) {
    chomp $line;

    $line =~ s/^\s+//g;
    $line =~ s/\s+$//g;

    if ( $line eq "" ) { next; }

    # header
    # Cluster, Unclaimed disk capacity (GiB), Unclaimed storage_volumes, Claimed disk capacity(GiB), Claimed storage_volumes, Used storage-volume capacity (GiB), Used storage_volumes, Unexported volume capacity (GiB), Unexported volumes, Exported volume capacity (GiB), Exported volumes
    if ( $line =~ m/^Cluster, Unclaimed/ ) {
      @header = split( ", ", $line );
      next;
    }

    if ( $line =~ m/^$cluster_id, / ) {
      my @values = split( ", ", $line );
      my $value_idx = 0;
      foreach my $value (@values) {
        chomp $value;
        $cap_data{ $header[$value_idx] } = $value;
        $value_idx++;
      }
    }
  }

  open( CAP, ">$capacity_file" ) || error( "Couldn't open file $capacity_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
  foreach my $key (keys %cap_data) {
    my $value = $cap_data{$key};
    print CAP "$key=$value\n";
  }
  close(CAP);

  return 1;
}

sub health_status {
  my $cmd_status_file = "$monitors_dir/cluster_status_commandresult";

  if ( -f $cmd_status_file ) {
    # check last update
    my $time_diff = file_time_diff($cmd_status_file);
    if ( $time_diff > 7200 ) {
      unlink("$cmd_status_file");
      error( "Something wrong with \"cluster status\" cmd! This file is older than 2 hours, this should not happen! \"$cmd_status_file\" was removed...  $!" . __FILE__ . ":" . __LINE__ ) && return;
    }

    open( STAT, "<$cmd_status_file" ) || error( "Couldn't open file $cmd_status_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
    my $url4cmdstatus = <STAT>;
    chomp $url4cmdstatus;
    close(STAT);

    #check URL
    if ( $url4cmdstatus !~ m/\/vplex\/commandresult\// ) {
      error( "health status : something wrong with this URL: \"$url4cmdstatus\" $!" . __FILE__ . ":" . __LINE__ );
      error( "health status : \"$cmd_status_file\" file will be removed! $!" . __FILE__ . ":" . __LINE__ );
      unlink("$cmd_status_file");
    }

    #check status
    my $result = new_get_request_configdump("$url4cmdstatus");
    if ( $result =~ m/Command execution continues to be under progress/ ) {    # still working..
      print_info("health status : Command execution continues to be under progress");
    }
    else {
      if ( defined $result->{'response'}{'custom-data'} && $result->{'response'}{'custom-data'} =~ m/Cluster $cluster_id/ ) {    # file is finished
        print_info("health status : Command was completed successfully");
        #print Dumper $result;
        unlink("$cmd_status_file");

        # parse and print cmd output
        my @out = split("\n", $result->{'response'}{'custom-data'});
        print_health_status(\@out);
      }
    }
  }
  else {
    print_info("health status : run cmd via API...");
    my $json_b = "{\"args\":\"-c $cluster_id\"}";
    my $result = new_post_request_configdump( "$url/cluster+status", $json_b );

    #print Dumper $result;
    if ( $result =~ m/Command execution taking longer than 60 seconds/ ) {

      #print "$result\n";
      my ( undef, $cmd_status_url ) = split( ",", $result );
      chomp $cmd_status_url;

      open( STAT, ">$cmd_status_file" ) || error( "Couldn't open file $cmd_status_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
      print STAT "$cmd_status_url\n";
      close(STAT);
    } ## end if ( $result =~ m/Command execution taking longer than 60 seconds/)
    else {
      print_info("health status : Command was completed successfully");
      #print Dumper $result;
      unlink("$cmd_status_file");

      # parse and print cmd output
      my @out = split("\n", $result->{'response'}{'custom-data'});
      print_health_status(\@out);
    }
  } ## end else [ if ( -f $cfgdump_cmd_status_file)]

  return 1;
}

sub print_health_status {
  my $cmd_out = shift;
  my @data    = @{$cmd_out};

  my $oper_st   = "";
  my $health_st = "";

  ($oper_st)   = grep {/operational-status:\s+/} @data;
  ($health_st) = grep {/health-state:\s+/} @data;

  if ( defined $oper_st && $oper_st ne '' && defined $health_st && $health_st ne '' ) {
    $oper_st =~ s/operational-status:\s+//;
    $oper_st =~ s/^\s+//;
    $oper_st =~ s/\s+$//;
    $health_st =~ s/health-state:\s+//;
    $health_st =~ s/^\s+//;
    $health_st =~ s/\s+$//;

    my $act_timestamp  = time();
    my $main_state     = "";
    my $state_suffix   = "";
    my $component_name = $storage_name;

    $component_name =~ s/\s+//g;
    my $alert_status = 0;

    if ( $oper_st =~ m/ok/ && $health_st =~ m/ok/ ) {
      $main_state = "OK";
      $state_suffix = "ok";
    }
    else {
      $main_state = "NOT_OK";
      $state_suffix = "nok";
    }

    if ( defined $main_state && $main_state ne '' && defined $state_suffix && $state_suffix ne '' ) {
      if ( !-d "$tmpdir/health_status_summary" ) {
        mkdir( "$tmpdir/health_status_summary", 0755 ) || error( "Cannot mkdir $tmpdir/health_status_summary: $!" . __FILE__ . ":" . __LINE__ ) && exit;
      }
      if ( !-f "$tmpdir/health_status_summary/$component_name.ok" && !-f "$tmpdir/health_status_summary/$component_name.nok" ) {
        $alert_status = 1;
      }
      if ( -f "$tmpdir/health_status_summary/$component_name.ok" ) { unlink("$tmpdir/health_status_summary/$component_name.ok"); $alert_status = 1; }
      if ( -f "$tmpdir/health_status_summary/$component_name.nok" ) { unlink("$tmpdir/health_status_summary/$component_name.nok"); }

      if ( $alert_status == 1 && $state_suffix eq "nok" ) {
        ### alert health status
        if ( exists &Xorux_lib::health_status_alerting ) {
          Xorux_lib::health_status_alerting($component_name,\@data);
        }
      } ## end if ( $alert_status == ...)

      open( MAST, ">$tmpdir/health_status_summary/$component_name.$state_suffix" ) || error( "Couldn't open file $tmpdir/health_status_summary/$component_name.$state_suffix $!" . __FILE__ . ":" . __LINE__ ) && exit;
      print MAST "STORAGE : $storage_name : $main_state : $act_timestamp\n";
      close(MAST);
    }
  }
  else {
    error( "health status : operational-status or health-state was not found via REST API!" . __FILE__ . ":" . __LINE__ );
    print_info("health status : operational-status or health-state was not found via REST API!");
    print join("\n", @data);
  }


  # create health_status.html
  my $last_update        = localtime();

  open( HS, ">$health_status_html" ) || error( "Couldn't open file $health_status_html $!" . __FILE__ . ":" . __LINE__ ) && exit;

  foreach my $line (@data) {
    chomp $line;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    if ( $line eq "" ) { next; }

    print HS "<br>$line\n";
  }

  print HS "<br>\n";
  print HS "<br>\n";
  print HS "<br>Last update: $last_update\n";

  close(HS);

  return 1;
}

sub download_configdump {
  my $configdump_filename = shift;

  my $time_to_scp             = 0;
  my $cfgdump_cmd_status_file = "$monitors_dir/configdump_commandresult";

  if ( -f $cfgdump_cmd_status_file ) {

    # check last update
    my $time_diff = file_time_diff($cfgdump_cmd_status_file);
    if ( $time_diff > 43200 ) { # timeout after 6 hours
      unlink("$cfgdump_cmd_status_file");
      error( "Something wrong with \"creating cluster configdump cmd\"! This file is older than 12 hours, this should not happen! \"$cfgdump_cmd_status_file\" was removed...  $!" . __FILE__ . ":" . __LINE__ ) && return;
    }

    open( STAT, "<$cfgdump_cmd_status_file" ) || error( "Couldn't open file $cfgdump_cmd_status_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
    my $url4cmdstatus = <STAT>;
    chomp $url4cmdstatus;
    close(STAT);

    #check URL
    if ( $url4cmdstatus !~ m/\/vplex\/commandresult\// ) {
      error( "configdump : something wrong with this URL: \"$url4cmdstatus\" $!" . __FILE__ . ":" . __LINE__ );
      error( "configdump : \"$cfgdump_cmd_status_file\" file will be removed! $!" . __FILE__ . ":" . __LINE__ );
      unlink("$cfgdump_cmd_status_file");
    }

    #check status
    my $result = new_get_request_configdump("$url4cmdstatus");
    if ( $result =~ m/Command execution continues to be under progress/ ) {    # still working..
                                                                               #print "$result\n";
      print_info("CONF : download_configdump : Command execution continues to be under progress");
    }
    else {
      if ( defined $result->{'response'}{'custom-data'} && $result->{'response'}{'custom-data'} =~ m/Initiating cluster configdump/ ) {    # file is finished
        $time_to_scp = 1;
        unlink("$cfgdump_cmd_status_file");
        print_info("CONF : download_configdump : Command was completed successfully");
      }
    } ## end else [ if ( $result =~ m/Command execution continues to be under progress/)]
  } ## end if ( -f $cfgdump_cmd_status_file)
  else {                                                                                                                                   # run cmd via api for creating configdump file
    print_info("CONF : download_configdump : Creating $configdump_filename via rest api...");
    my $json_b = "{\"args\":\"-c $cluster_id -f /var/log/VPlex/cli/$configdump_filename\"}";
    my $result = new_post_request_configdump( "$url/cluster+configdump", $json_b );

    #print Dumper $result;
    if ( $result =~ m/Command execution taking longer than 60 seconds/ ) {

      #print "$result\n";
      my ( undef, $cmd_status_url ) = split( ",", $result );
      chomp $cmd_status_url;

      open( STAT, ">$cfgdump_cmd_status_file" ) || error( "Couldn't open file $cfgdump_cmd_status_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
      print STAT "$cmd_status_url\n";
      close(STAT);
    } ## end if ( $result =~ m/Command execution taking longer than 60 seconds/)
    else {
      $time_to_scp = 1;
    }
  } ## end else [ if ( -f $cfgdump_cmd_status_file)]

  # copy configdump.xml from mgmt server
  if ( $time_to_scp eq "1" ) {
    print_info("CONF : download_configdump : Copy $configdump_filename from mgmt server: $vplex_ssh_ip");
    copy_configdump("$configdump_filename","$vplex_ssh_ip");

    # try to copy configdump file from server where is Unisphere installed
    if ( ! -f "$monitors_dir/$configdump_filename" ) {
      print_info("CONF : download_configdump : Copy $configdump_filename from mgmt server where Unisphere is istalled: $vplex_api_ip");
      copy_configdump("$configdump_filename","$vplex_api_ip");
    }

    if ( -f "$monitors_dir/$configdump_filename" ) {
      return 1;
    }
    else {
      error( "$monitors_dir/$configdump_filename not exists! Something wrong with copy procedure... $!" . __FILE__ . ":" . __LINE__ ) && return 0;
    }
  } ## end if ( $time_to_scp eq "1")

  return 0;
} ## end sub download_configdump

sub copy_configdump {
  my $configdump_filename = shift;
  my $act_ip              = shift;

  #my $file_line = `$SSH $vplex_ssh_user\@$vplex_ssh_ip -p $vplex_ssh_port "LC_ALL=en_US LANG=en_US LC_NUMERIC=en_US ls -l /var/log/VPlex/cli/$configdump_filename" 2>\&1`;

  #my (undef, undef, undef, undef, $size, undef, undef, undef, $file) = split(/\s+/, $line);
  #my $filename = $file;
  #$filename =~ s/^\/var\/log\/VPlex\/cli\///;

  my $scp_cmd = "scp -P $vplex_ssh_port $vplex_ssh_user\@$act_ip:/var/log/VPlex/cli/$configdump_filename $monitors_dir";

  #print "$scp_cmd\n";

  my $scp_out = `$scp_cmd`;

  #print "$scp_out\n";

  return 1;
} ## end sub copy_configdump

sub check_clusters {
  my $result = new_get_request("$url/clusters");
  if ( $result eq "0" ) { return 0; }

  if ( defined $result->{'response'}{'context'} && $result->{'response'}{'context'} ne '' ) {
    foreach ( @{ $result->{'response'}{'context'} } ) {
      if ( defined $_->{'children'} && $_->{'children'} ne "" ) {
        foreach ( @{ $_->{'children'} } ) {
          if ( defined $_->{'name'} && $_->{'name'} ne "" ) {

            #print "$_->{'name'}\n";
            $data{'conf'}{'cluster'}{ $_->{'name'} }{'name'} = $_->{'name'};
          }
        } ## end foreach ( @{ $_->{'children'...}})
      } ## end if ( defined $_->{'children'...})
    } ## end foreach ( @{ $result->{'response'...}})
  } ## end if ( defined $result->...)

  return 1;
} ## end sub check_clusters

sub print_perf {
  open( PERF, ">$perf_out" ) || error( "Couldn't open file $perf_out $!" . __FILE__ . ":" . __LINE__ ) && exit;

  # Virtual volumes
  print PERF "Volume Level Statistics\n";
  print PERF "---------------------\n";
  print PERF "Virtual Volume,timestamp,date,director,io_rate,read (KB/s),write (KB/s),resp_t_r (ms),resp_t_w (ms),data_rate (KB/s)\n";

  my $vvol_idx = 0;
  foreach my $vvol ( sort keys( %{ $data{'perf'}{'virtual volumes'} } ) ) {

    #print "   $dir\n";
    $vvol_idx++;
    if ( defined $data{'perf'}{'virtual volumes'}{$vvol} && $data{'perf'}{'virtual volumes'}{$vvol} ne '' ) {
      my $dir_idx           = 0;
      my $last_printed_line = "";

      my $io_rate   = "";
      my $read      = "";
      my $write     = "";
      my $data_rate = "";
      my $resp_t_r  = "";
      my $resp_t_w  = "";

      my $resp_t_r_idx = 0;
      my $resp_t_w_idx = 0;

      my $timestamp = "";
      my $date      = "";
      my $directors = "";

      foreach my $dir ( sort keys( %{ $data{'perf'}{'virtual volumes'}{$vvol} } ) ) {

        #print "     $vvol\n";
        $dir_idx++;

        my $io_rate_act = "";
        my $read_act    = "";
        my $write_act   = "";

        my $io_rate_idx = 0;
        my $read_idx    = 0;
        my $write_idx   = 0;

        if ( defined $directors && $directors ne '' ) {
          $directors .= " $dir";
        }
        else {
          $directors = "$dir";
        }

        foreach my $time_r ( sort keys( %{ $data{'perf'}{'virtual volumes'}{$vvol}{$dir} } ) ) {

          # io_rate = fe-lu ops (counts/s)
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu ops (counts/s)'} && isdigit( $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu ops (counts/s)'} ) ) {
            if ( defined $io_rate_act && isdigit($io_rate_act) ) {
              $io_rate_act += $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu ops (counts/s)'};
            }
            else {
              $io_rate_act = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu ops (counts/s)'};
            }
            $io_rate_idx++;
          } ## end if ( defined $data{'perf'...})

          # read = fe-lu read (KB/s)
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read (KB/s)'} && isdigit( $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read (KB/s)'} ) ) {
            if ( defined $read_act && isdigit($read_act) ) {
              $read_act += $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read (KB/s)'};
            }
            else {
              $read_act = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read (KB/s)'};
            }
            $read_idx++;
          } ## end if ( defined $data{'perf'...})

          # write = fe-lu write (KB/s)
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write (KB/s)'} && isdigit( $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write (KB/s)'} ) ) {
            if ( defined $write_act && isdigit($write_act) ) {
              $write_act += $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write (KB/s)'};
            }
            else {
              $write_act = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write (KB/s)'};
            }
            $write_idx++;
          } ## end if ( defined $data{'perf'...})

          # resp_t_r = fe-lu read-lat recent-average (us)
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read-lat recent-average (us)'} && isdigit( $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read-lat recent-average (us)'} ) ) {
            my $resp_t_r_act = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read-lat recent-average (us)'} / 1000;    # from us to ms
            if ( defined $resp_t_r && isdigit($resp_t_r) ) {
              $resp_t_r += $resp_t_r_act;
            }
            else {
              $resp_t_r = $resp_t_r_act;
            }
            $resp_t_r_idx++;
          } ## end if ( defined $data{'perf'...})

          # resp_r_w = fe-lu write-lat recent-average (us)
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write-lat recent-average (us)'} && isdigit( $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write-lat recent-average (us)'} ) ) {
            my $resp_t_w_act = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write-lat recent-average (us)'} / 1000;    # from us to ms
            if ( defined $resp_t_w && isdigit($resp_t_w) ) {
              $resp_t_w += $resp_t_w_act;
            }
            else {
              $resp_t_w = $resp_t_w_act;
            }
            $resp_t_w_idx++;
          } ## end if ( defined $data{'perf'...})
          #
          # vplex firmware version 5.3.0.03.00.04 has got different names of resp_t statistics
          # resp_t_r = fe-lu read-avglat (us)
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read-avglat (us)'} && isdigit( $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read-avglat (us)'} ) ) {
            my $resp_t_r_act = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu read-avglat (us)'} / 1000;    # from us to ms
            if ( defined $resp_t_r && isdigit($resp_t_r) ) {
              $resp_t_r += $resp_t_r_act;
            }
            else {
              $resp_t_r = $resp_t_r_act;
            }
            $resp_t_r_idx++;
          } ## end if ( defined $data{'perf'...})

          # resp_r_w = fe-lu write-avglat (us)
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write-avglat (us)'} && isdigit( $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write-avglat (us)'} ) ) {
            my $resp_t_w_act = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'fe-lu write-avglat (us)'} / 1000;    # from us to ms
            if ( defined $resp_t_w && isdigit($resp_t_w) ) {
              $resp_t_w += $resp_t_w_act;
            }
            else {
              $resp_t_w = $resp_t_w_act;
            }
            $resp_t_w_idx++;
          } ## end if ( defined $data{'perf'...})
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'LAST_TIMESTAMP'} && isdigit( $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'LAST_TIMESTAMP'} ) ) {
            $timestamp = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'LAST_TIMESTAMP'};
          }
          if ( defined $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'LAST_TIME'} && $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'LAST_TIME'} ne '' ) {
            $date = $data{'perf'}{'virtual volumes'}{$vvol}{$dir}{$time_r}{'LAST_TIME'};
          }
        } ## end foreach my $time_r ( sort keys...)
        if ( defined $io_rate_act && isdigit($io_rate_act) && defined $io_rate_idx && isdigit($io_rate_idx) && $io_rate_idx ne "0" ) {
          $io_rate_act = $io_rate_act / $io_rate_idx;
          $io_rate_act = sprintf( '%.0f', $io_rate_act );
          if ( defined $io_rate && isdigit($io_rate) ) {
            $io_rate += $io_rate_act;
          }
          else {
            $io_rate = $io_rate_act;
          }
        } ## end if ( defined $io_rate_act...)
        if ( defined $read_act && isdigit($read_act) && defined $read_idx && isdigit($read_idx) && $read_idx ne "0" ) {
          $read_act = $read_act / $read_idx;
          $read_act = sprintf( '%.2f', $read_act );
          if ( defined $read && isdigit($read) ) {
            $read += $read_act;
          }
          else {
            $read = $read_act;
          }
        } ## end if ( defined $read_act...)
        if ( defined $write_act && isdigit($write_act) && defined $write_idx && isdigit($write_idx) && $write_idx ne "0" ) {
          $write_act = $write_act / $write_idx;
          $write_act = sprintf( '%.2f', $write_act );
          if ( defined $write && isdigit($write) ) {
            $write += $write_act;
          }
          else {
            $write = $write_act;
          }
        } ## end if ( defined $write_act...)
      } ## end foreach my $dir ( sort keys...)

      if ( defined $resp_t_r && isdigit($resp_t_r) && defined $resp_t_r_idx && isdigit($resp_t_r_idx) && $resp_t_r_idx ne "0" ) {
        $resp_t_r = $resp_t_r / $resp_t_r_idx;
      }
      if ( defined $resp_t_w && isdigit($resp_t_w) && defined $resp_t_w_idx && isdigit($resp_t_w_idx) && $resp_t_w_idx ne "0" ) {
        $resp_t_w = $resp_t_w / $resp_t_w_idx;
      }
      if ( defined $read && isdigit($read) && defined $write && isdigit($write) ) {
        $data_rate = $read + $write;
      }

      print PERF "$vvol,$timestamp,$date,$directors,$io_rate,$read,$write,$resp_t_r,$resp_t_w,$data_rate\n";
    } ## end if ( defined $data{'perf'...})
  } ## end foreach my $vvol ( sort keys...)
  print PERF "\n";

  # Director stats (NODE)
  print PERF "Director Level Statistics\n";
  print PERF "---------------------\n";
  print PERF "Director,timestamp,date,busy (%),io_rate_b,read_io_b,write_io_b,read_b (KB/s),write_b (KB/s),io_rate,read_io,write_io,read (KB/s),write (KB/s),data_rate_b (KB/s),data_rate (KB/s),resp_t_r,resp_t_w\n";

  foreach my $dir ( sort keys( %{ $data{'perf'} } ) ) {

    #print "   $dir\n";

    if ( $dir eq "virtual volumes" ) { next; }

    my $busy_prct   = "";
    my $io_rate_b   = "";
    my $read_io_b   = "";
    my $write_io_b  = "";
    my $read_b      = "";
    my $write_b     = "";
    my $io_rate     = "";
    my $read_io     = "";
    my $write_io    = "";
    my $read        = "";
    my $write       = "";
    my $data_rate_b = "";
    my $data_rate   = "";
    my $resp_t_r    = "";
    my $resp_t_w    = "";

    my $busy_prct_idx  = 0;
    my $io_rate_b_idx  = 0;
    my $read_io_b_idx  = 0;
    my $write_io_b_idx = 0;
    my $read_b_idx     = 0;
    my $write_b_idx    = 0;
    my $io_rate_idx    = 0;
    my $read_io_idx    = 0;
    my $write_io_idx   = 0;
    my $read_idx       = 0;
    my $write_idx      = 0;
    my $resp_t_r_idx   = 0;
    my $resp_t_w_idx   = 0;

    my $timestamp = "";
    my $date      = "";

    if ( defined $data{'perf'}{$dir}{'director'} && $data{'perf'}{$dir}{'director'} ne '' ) {
      foreach my $time_r ( sort keys( %{ $data{'perf'}{$dir}{'director'} } ) ) {

        # busy = director.busy (%)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.busy (%)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.busy (%)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.busy (%)'}\n";
          if ( defined $busy_prct && isdigit($busy_prct) ) {
            $busy_prct += $data{'perf'}{$dir}{'director'}{$time_r}{'director.busy (%)'};
          }
          else {
            $busy_prct = $data{'perf'}{$dir}{'director'}{$time_r}{'director.busy (%)'};
          }
          $busy_prct_idx++;
        } ## end if ( defined $data{'perf'...})

        # io_rate_b = director.be-ops (counts/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops (counts/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops (counts/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops (counts/s)'}\n";
          if ( defined $io_rate_b && isdigit($io_rate_b) ) {
            $io_rate_b += $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops (counts/s)'};
          }
          else {
            $io_rate_b = $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops (counts/s)'};
          }
          $io_rate_b_idx++;
        } ## end if ( defined $data{'perf'...})

        # read_io_b = director.be-ops-read (counts/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-read (counts/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-read (counts/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-read (counts/s)'}\n";
          if ( defined $read_io_b && isdigit($read_io_b) ) {
            $read_io_b += $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-read (counts/s)'};
          }
          else {
            $read_io_b = $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-read (counts/s)'};
          }
          $read_io_b_idx++;
        } ## end if ( defined $data{'perf'...})

        # write_io_b = director.be-ops-write (counts/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-write (counts/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-write (counts/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-write (counts/s)'}\n";
          if ( defined $write_io_b && isdigit($write_io_b) ) {
            $write_io_b += $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-write (counts/s)'};
          }
          else {
            $write_io_b = $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-ops-write (counts/s)'};
          }
          $write_io_b_idx++;
        } ## end if ( defined $data{'perf'...})

        # read_b = director.be-read (KB/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-read (KB/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-read (KB/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-read (KB/s)'}\n";
          if ( defined $read_b && isdigit($read_b) ) {
            $read_b += $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-read (KB/s)'};
          }
          else {
            $read_b = $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-read (KB/s)'};
          }
          $read_b_idx++;
        } ## end if ( defined $data{'perf'...})

        # write_b = director.be-write (KB/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-write (KB/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-write (KB/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-write (KB/s)'}\n";
          if ( defined $write_b && isdigit($write_b) ) {
            $write_b += $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-write (KB/s)'};
          }
          else {
            $write_b = $data{'perf'}{$dir}{'director'}{$time_r}{'director.be-write (KB/s)'};
          }
          $write_b_idx++;
        } ## end if ( defined $data{'perf'...})

        # io_rate = director.fe-ops (counts/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops (counts/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops (counts/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops (counts/s)'}\n";
          if ( defined $io_rate && isdigit($io_rate) ) {
            $io_rate += $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops (counts/s)'};
          }
          else {
            $io_rate = $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops (counts/s)'};
          }
          $io_rate_idx++;
        } ## end if ( defined $data{'perf'...})

        # read_io = director.fe-ops-read (counts/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-read (counts/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-read (counts/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-read (counts/s)'}\n";
          if ( defined $read_io && isdigit($read_io) ) {
            $read_io += $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-read (counts/s)'};
          }
          else {
            $read_io = $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-read (counts/s)'};
          }
          $read_io_idx++;
        } ## end if ( defined $data{'perf'...})

        # write_io = director.fe-ops-write (counts/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-write (counts/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-write (counts/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-write (counts/s)'}\n";
          if ( defined $write_io && isdigit($write_io) ) {
            $write_io += $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-write (counts/s)'};
          }
          else {
            $write_io = $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-ops-write (counts/s)'};
          }
          $write_io_idx++;
        } ## end if ( defined $data{'perf'...})

        # read = director.fe-read (KB/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-read (KB/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-read (KB/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-read (KB/s)'}\n";
          if ( defined $read && isdigit($read) ) {
            $read += $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-read (KB/s)'};
          }
          else {
            $read = $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-read (KB/s)'};
          }
          $read_idx++;
        } ## end if ( defined $data{'perf'...})

        # write = director.fe-write (KB/s)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-write (KB/s)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-write (KB/s)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-write (KB/s)'}\n";
          if ( defined $write && isdigit($write) ) {
            $write += $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-write (KB/s)'};
          }
          else {
            $write = $data{'perf'}{$dir}{'director'}{$time_r}{'director.fe-write (KB/s)'};
          }
          $write_idx++;
        } ## end if ( defined $data{'perf'...})

        # resp_t_r = fe-director.read-lat recent-average (us)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.read-lat recent-average (us)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.read-lat recent-average (us)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.read-lat recent-average (us)'}\n";
          if ( defined $resp_t_r && isdigit($resp_t_r) ) {
            $resp_t_r += $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.read-lat recent-average (us)'};
          }
          else {
            $resp_t_r = $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.read-lat recent-average (us)'};
          }
          $resp_t_r_idx++;
        } ## end if ( defined $data{'perf'...})

        # resp_t_w = fe-director.write-lat recent-average (us)
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.write-lat recent-average (us)'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.write-lat recent-average (us)'} ) ) {

          #print "  $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.write-lat recent-average (us)'}\n";
          if ( defined $resp_t_w && isdigit($resp_t_w) ) {
            $resp_t_w += $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.write-lat recent-average (us)'};
          }
          else {
            $resp_t_w = $data{'perf'}{$dir}{'director'}{$time_r}{'fe-director.write-lat recent-average (us)'};
          }
          $resp_t_w_idx++;
        } ## end if ( defined $data{'perf'...})
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIMESTAMP'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIMESTAMP'} ) ) {
          $timestamp = $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIMESTAMP'};
        }
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIME'} && $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIME'} ne '' ) {
          $date = $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIME'};
        }
      } ## end foreach my $time_r ( sort keys...)
    } ## end if ( defined $data{'perf'...})

    if ( defined $busy_prct && isdigit($busy_prct) && defined $busy_prct_idx && isdigit($busy_prct_idx) && $busy_prct_idx ne "0" ) {
      $busy_prct = $busy_prct / $busy_prct_idx;
      $busy_prct = sprintf( '%.0f', $busy_prct );
    }
    if ( defined $io_rate_b && isdigit($io_rate_b) && defined $io_rate_b_idx && isdigit($io_rate_b_idx) && $io_rate_b_idx ne "0" ) {
      $io_rate_b = $io_rate_b / $io_rate_b_idx;
      $io_rate_b = sprintf( '%.0f', $io_rate_b );
    }
    if ( defined $read_io_b && isdigit($read_io_b) && defined $read_io_b_idx && isdigit($read_io_b_idx) && $read_io_b_idx ne "0" ) {
      $read_io_b = $read_io_b / $read_io_b_idx;
      $read_io_b = sprintf( '%.0f', $read_io_b );
    }
    if ( defined $write_io_b && isdigit($write_io_b) && defined $write_io_b_idx && isdigit($write_io_b_idx) && $write_io_b_idx ne "0" ) {
      $write_io_b = $write_io_b / $write_io_b_idx;
      $write_io_b = sprintf( '%.0f', $write_io_b );
    }
    if ( defined $read_b && isdigit($read_b) && defined $read_b_idx && isdigit($read_b_idx) && $read_b_idx ne "0" ) {
      $read_b = $read_b / $read_b_idx;
      $read_b = sprintf( '%.2f', $read_b );
    }
    if ( defined $write_b && isdigit($write_b) && defined $write_b_idx && isdigit($write_b_idx) && $write_b_idx ne "0" ) {
      $write_b = $write_b / $write_b_idx;
      $write_b = sprintf( '%.2f', $write_b );
    }
    if ( defined $io_rate && isdigit($io_rate) && defined $io_rate_idx && isdigit($io_rate_idx) && $io_rate_idx ne "0" ) {
      $io_rate = $io_rate / $io_rate_idx;
      $io_rate = sprintf( '%.0f', $io_rate );
    }
    if ( defined $read_io && isdigit($read_io) && defined $read_io_idx && isdigit($read_io_idx) && $read_io_idx ne "0" ) {
      $read_io = $read_io / $read_io_idx;
      $read_io = sprintf( '%.0f', $read_io );
    }
    if ( defined $write_io && isdigit($write_io) && defined $write_io_idx && isdigit($write_io_idx) && $write_io_idx ne "0" ) {
      $write_io = $write_io / $write_io_idx;
      $write_io = sprintf( '%.0f', $write_io );
    }
    if ( defined $read && isdigit($read) && defined $read_idx && isdigit($read_idx) && $read_idx ne "0" ) {
      $read = $read / $read_idx;
      $read = sprintf( '%.2f', $read );
    }
    if ( defined $write && isdigit($write) && defined $write_idx && isdigit($write_idx) && $write_idx ne "0" ) {
      $write = $write / $write_idx;
      $write = sprintf( '%.2f', $write );
    }
    if ( defined $resp_t_r && isdigit($resp_t_r) && defined $resp_t_r_idx && isdigit($resp_t_r_idx) && $resp_t_r_idx ne "0" ) {
      $resp_t_r = $resp_t_r / $resp_t_r_idx;
      $resp_t_r = sprintf( '%.2f', $resp_t_r / 1000 );
    }
    if ( defined $resp_t_w && isdigit($resp_t_w) && defined $resp_t_w_idx && isdigit($resp_t_w_idx) && $resp_t_w_idx ne "0" ) {
      $resp_t_w = $resp_t_w / $resp_t_w_idx;
      $resp_t_w = sprintf( '%.2f', $resp_t_w / 1000 );
    }

    if ( defined $read_b && isdigit($read_b) && defined $write_b && isdigit($write_b) ) {
      $data_rate_b = $read_b + $write_b;
    }
    if ( defined $read && isdigit($read) && defined $write && isdigit($write) ) {
      $data_rate = $read + $write;
    }

    print PERF "$dir,$timestamp,$date,$busy_prct,$io_rate_b,$read_io_b,$write_io_b,$read_b,$write_b,$io_rate,$read_io,$write_io,$read,$write,$data_rate_b,$data_rate,$resp_t_r,$resp_t_w\n";
  } ## end foreach my $dir ( sort keys...)
  print PERF "\n";

  # BEPorts and FEPorts stats
  # parse ports
  my %ports;
  foreach my $dir ( sort keys( %{ $data{'perf'} } ) ) {

    #print "   $dir\n";

    if ( defined $data{'perf'}{$dir}{'director'} && $data{'perf'}{$dir}{'director'} ne '' ) {
      foreach my $time_r ( sort keys( %{ $data{'perf'}{$dir}{'director'} } ) ) {

        # set last update time
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIMESTAMP'} && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIMESTAMP'} ) ) {
          $ports{'director'}{$dir}{'timestamp'} = $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIMESTAMP'};
        }
        if ( defined $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIME'} && $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIME'} ne '' ) {
          $ports{'director'}{$dir}{'date'} = $data{'perf'}{$dir}{'director'}{$time_r}{'LAST_TIME'};
        }

        # search for port metric
        foreach my $metric ( sort keys( %{ $data{'perf'}{$dir}{'director'}{$time_r} } ) ) {

          # BEPorts read
          if ( $metric =~ m/^be-prt.read / && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{$metric} ) ) {
            my ( undef, $port_name, undef ) = split( /\s+/, $metric );

            #print "   $port_name : $metric\n";

            if ( defined $ports{'director'}{$dir}{'BEport'}{$port_name}{'read'} && isdigit( $ports{'director'}{$dir}{'BEport'}{$port_name}{'read'} ) ) {
              $ports{'director'}{$dir}{'BEport'}{$port_name}{'read'} += $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'BEport'}{$port_name}{'read calculate'} .= " + $data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
            else {
              $ports{'director'}{$dir}{'BEport'}{$port_name}{'read'}           = $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'BEport'}{$port_name}{'read calculate'} = "$data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
          } ## end if ( $metric =~ m/^be-prt.read /...)

          # BEPorts write
          if ( $metric =~ m/^be-prt.write / && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{$metric} ) ) {
            my ( undef, $port_name, undef ) = split( /\s+/, $metric );

            #print "   $port_name : $metric\n";

            if ( defined $ports{'director'}{$dir}{'BEport'}{$port_name}{'write'} && isdigit( $ports{'director'}{$dir}{'BEport'}{$port_name}{'write'} ) ) {
              $ports{'director'}{$dir}{'BEport'}{$port_name}{'write'} += $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'BEport'}{$port_name}{'write calculate'} .= " + $data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
            else {
              $ports{'director'}{$dir}{'BEport'}{$port_name}{'write'}           = $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'BEport'}{$port_name}{'write calculate'} = "$data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
          } ## end if ( $metric =~ m/^be-prt.write /...)

          # FEPorts read
          if ( $metric =~ m/^fe-prt.read / && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{$metric} ) ) {
            my ( undef, $port_name, undef ) = split( /\s+/, $metric );

            #print "   $port_name : $metric\n";

            if ( defined $ports{'director'}{$dir}{'FEport'}{$port_name}{'read'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port_name}{'read'} ) ) {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'read'} += $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'read calculate'} .= " + $data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
            else {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'read'}           = $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'read calculate'} = "$data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
          } ## end if ( $metric =~ m/^fe-prt.read /...)

          # FEPorts write
          if ( $metric =~ m/^fe-prt.write / && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{$metric} ) ) {
            my ( undef, $port_name, undef ) = split( /\s+/, $metric );

            #print "   $port_name : $metric\n";

            if ( defined $ports{'director'}{$dir}{'FEport'}{$port_name}{'write'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port_name}{'write'} ) ) {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'write'} += $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'write calculate'} .= " + $data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
            else {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'write'}           = $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'write calculate'} = "$data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
          } ## end if ( $metric =~ m/^fe-prt.write /...)

          # FEPorts io_rate
          if ( $metric =~ m/^fe-prt.ops / && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{$metric} ) ) {
            my ( undef, $port_name, undef ) = split( /\s+/, $metric );

            #print "   $port_name : $metric\n";

            if ( defined $ports{'director'}{$dir}{'FEport'}{$port_name}{'io_rate'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port_name}{'io_rate'} ) ) {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'io_rate'} += $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'io_rate calculate'} .= " + $data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
            else {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'io_rate'}           = $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'io_rate calculate'} = "$data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
          } ## end if ( $metric =~ m/^fe-prt.ops /...)

          # FEPorts resp_t_r
          if ( $metric =~ m/^fe-prt.read-avg-lat / && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{$metric} ) ) {
            my ( undef, $port_name, undef ) = split( /\s+/, $metric );

            #print "   $port_name : $metric\n";

            if ( defined $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_r'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_r'} ) ) {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_r'} += $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_r calculate'} .= " + $data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
            else {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_r'}           = $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_r calculate'} = "$data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
          } ## end if ( $metric =~ m/^fe-prt.read-avg-lat /...)

          # FEPorts resp_t_w
          if ( $metric =~ m/^fe-prt.write-avg-lat / && isdigit( $data{'perf'}{$dir}{'director'}{$time_r}{$metric} ) ) {
            my ( undef, $port_name, undef ) = split( /\s+/, $metric );

            #print "   $port_name : $metric\n";

            if ( defined $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_w'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_w'} ) ) {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_w'} += $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_w calculate'} .= " + $data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
            else {
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_w'}           = $data{'perf'}{$dir}{'director'}{$time_r}{$metric};
              $ports{'director'}{$dir}{'FEport'}{$port_name}{'resp_t_w calculate'} = "$data{'perf'}{$dir}{'director'}{$time_r}{$metric}";
            }
          } ## end if ( $metric =~ m/^fe-prt.write-avg-lat /...)
        } ## end foreach my $metric ( sort keys...)
      } ## end foreach my $time_r ( sort keys...)
    } ## end if ( defined $data{'perf'...})
  } ## end foreach my $dir ( sort keys...)

  # print stats
  # BEPort
  print PERF "BEPort Level Statistics\n";
  print PERF "---------------------\n";
  print PERF "Director,Port,timestamp,date,read (KB/s),write (KB/s),data_rate (KB/s),io_rate,resp_t_r,resp_t_w\n";

  foreach my $dir ( sort keys( %{ $ports{'director'} } ) ) {
    if ( defined $ports{'director'}{$dir}{'BEport'} && $ports{'director'}{$dir}{'BEport'} ne '' ) {
      foreach my $port ( sort keys( %{ $ports{'director'}{$dir}{'BEport'} } ) ) {
        my $read      = "";
        my $write     = "";
        my $data_rate = "";
        my $timestamp = "";
        my $date      = "";

        if ( defined $ports{'director'}{$dir}{'BEport'}{$port}{'read'} && isdigit( $ports{'director'}{$dir}{'BEport'}{$port}{'read'} ) ) {
          if ( defined $ports{'director'}{$dir}{'BEport'}{$port}{'read calculate'} && $ports{'director'}{$dir}{'BEport'}{$port}{'read calculate'} ne '' ) {
            my @nums_arr = split( / \+ /, $ports{'director'}{$dir}{'BEport'}{$port}{'read calculate'} );
            my $nums = @nums_arr;
            if ( defined $nums && isdigit($nums) && $nums ne "0" ) {
              $read = $ports{'director'}{$dir}{'BEport'}{$port}{'read'} / $nums;
              $read = sprintf( '%.2f', $read );
            }
          } ## end if ( defined $ports{'director'...})
        } ## end if ( defined $ports{'director'...})
        if ( defined $ports{'director'}{$dir}{'BEport'}{$port}{'write'} && isdigit( $ports{'director'}{$dir}{'BEport'}{$port}{'write'} ) ) {
          if ( defined $ports{'director'}{$dir}{'BEport'}{$port}{'write calculate'} && $ports{'director'}{$dir}{'BEport'}{$port}{'write calculate'} ne '' ) {
            my @nums_arr = split( / \+ /, $ports{'director'}{$dir}{'BEport'}{$port}{'write calculate'} );
            my $nums = @nums_arr;
            if ( defined $nums && isdigit($nums) && $nums ne "0" ) {
              $write = $ports{'director'}{$dir}{'BEport'}{$port}{'write'} / $nums;
              $write = sprintf( '%.2f', $write );
            }
          } ## end if ( defined $ports{'director'...})
        } ## end if ( defined $ports{'director'...})
        if ( defined $ports{'director'}{$dir}{'timestamp'} && isdigit( $ports{'director'}{$dir}{'timestamp'} ) ) {
          $timestamp = $ports{'director'}{$dir}{'timestamp'};
        }
        if ( defined $ports{'director'}{$dir}{'date'} && $ports{'director'}{$dir}{'date'} ne '' ) {
          $date = $ports{'director'}{$dir}{'date'};
        }
        if ( defined $read && isdigit($read) && defined $write && isdigit($write) ) {
          $data_rate = $read + $write;
        }

        print PERF "$dir,$port,$timestamp,$date,$read,$write,$data_rate,,,\n";
      } ## end foreach my $port ( sort keys...)
    } ## end if ( defined $ports{'director'...})
  } ## end foreach my $dir ( sort keys...)
  print PERF "\n";

  # print stats
  # FEPort
  print PERF "FEPort Level Statistics\n";
  print PERF "---------------------\n";
  print PERF "Director,Port,timestamp,date,read (KB/s),write (KB/s),data_rate (KB/s),io_rate,resp_t_r,resp_t_w\n";

  foreach my $dir ( sort keys( %{ $ports{'director'} } ) ) {
    if ( defined $ports{'director'}{$dir}{'FEport'} && $ports{'director'}{$dir}{'FEport'} ne '' ) {
      foreach my $port ( sort keys( %{ $ports{'director'}{$dir}{'FEport'} } ) ) {
        my $read      = "";
        my $write     = "";
        my $data_rate = "";
        my $io_rate   = "";
        my $resp_t_r  = "";
        my $resp_t_w  = "";
        my $timestamp = "";
        my $date      = "";

        if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'read'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port}{'read'} ) && $ports{'director'}{$dir}{'FEport'}{$port}{'read'} ne "0" ) {
          if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'read calculate'} && $ports{'director'}{$dir}{'FEport'}{$port}{'read calculate'} ne '' ) {
            my @nums_arr = split( / \+ /, $ports{'director'}{$dir}{'FEport'}{$port}{'read calculate'} );
            my $nums = @nums_arr;
            if ( defined $nums && isdigit($nums) && $nums ne "0" ) {
              $read = $ports{'director'}{$dir}{'FEport'}{$port}{'read'} / $nums;
              $read = sprintf( '%.2f', $read );
            }
          } ## end if ( defined $ports{'director'...})
        } ## end if ( defined $ports{'director'...})
        if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'write'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port}{'write'} ) && $ports{'director'}{$dir}{'FEport'}{$port}{'write'} ne "0" ) {
          if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'write calculate'} && $ports{'director'}{$dir}{'FEport'}{$port}{'write calculate'} ne '' ) {
            my @nums_arr = split( / \+ /, $ports{'director'}{$dir}{'FEport'}{$port}{'write calculate'} );
            my $nums = @nums_arr;
            if ( defined $nums && isdigit($nums) && $nums ne "0" ) {
              $write = $ports{'director'}{$dir}{'FEport'}{$port}{'write'} / $nums;
              $write = sprintf( '%.0f', $write );
            }
          } ## end if ( defined $ports{'director'...})
        } ## end if ( defined $ports{'director'...})
        if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'io_rate'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port}{'io_rate'} ) && $ports{'director'}{$dir}{'FEport'}{$port}{'io_rate'} ne "0" ) {
          if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'io_rate calculate'} && $ports{'director'}{$dir}{'FEport'}{$port}{'io_rate calculate'} ne '' ) {
            my @nums_arr = split( / \+ /, $ports{'director'}{$dir}{'FEport'}{$port}{'io_rate calculate'} );
            my $nums = @nums_arr;
            if ( defined $nums && isdigit($nums) && $nums ne "0" ) {
              $io_rate = $ports{'director'}{$dir}{'FEport'}{$port}{'io_rate'} / $nums;
              $io_rate = sprintf( '%.0f', $io_rate );
            }
          } ## end if ( defined $ports{'director'...})
        } ## end if ( defined $ports{'director'...})
        if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_r'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_r'} ) && $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_r'} ne "0" ) {
          if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_r calculate'} && $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_r calculate'} ne '' ) {
            my @nums_arr = split( / \+ /, $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_r calculate'} );
            my $nums = @nums_arr;
            if ( defined $nums && isdigit($nums) && $nums ne "0" ) {
              $resp_t_r = $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_r'} / $nums;
              $resp_t_r = sprintf( '%.2f', $resp_t_r / 1000 );
            }
          } ## end if ( defined $ports{'director'...})
        } ## end if ( defined $ports{'director'...})
        if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_w'} && isdigit( $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_w'} ) && $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_w'} ne "0" ) {
          if ( defined $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_w calculate'} && $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_w calculate'} ne '' ) {
            my @nums_arr = split( / \+ /, $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_w calculate'} );
            my $nums = @nums_arr;
            if ( defined $nums && isdigit($nums) && $nums ne "0" ) {
              $resp_t_w = $ports{'director'}{$dir}{'FEport'}{$port}{'resp_t_w'} / $nums;
              $resp_t_w = sprintf( '%.2f', $resp_t_w / 1000 );
            }
          } ## end if ( defined $ports{'director'...})
        } ## end if ( defined $ports{'director'...})
        if ( defined $ports{'director'}{$dir}{'timestamp'} && isdigit( $ports{'director'}{$dir}{'timestamp'} ) ) {
          $timestamp = $ports{'director'}{$dir}{'timestamp'};
        }
        if ( defined $ports{'director'}{$dir}{'date'} && $ports{'director'}{$dir}{'date'} ne '' ) {
          $date = $ports{'director'}{$dir}{'date'};
        }
        if ( defined $read && isdigit($read) && defined $write && isdigit($write) ) {
          $data_rate = $read + $write;
        }

        print PERF "$dir,$port,$timestamp,$date,$read,$write,$data_rate,$io_rate,$resp_t_r,$resp_t_w\n";
      } ## end foreach my $port ( sort keys...)
    } ## end if ( defined $ports{'director'...})
  } ## end foreach my $dir ( sort keys...)
  print PERF "\n";

  close(PERF);

  return 1;
} ## end sub print_perf

sub parse_monitors {
  if ( defined $monitors{DIRECTOR} && $monitors{DIRECTOR} ne '' ) {
    foreach my $director ( sort keys( %{ $monitors{DIRECTOR} } ) ) {
      foreach my $monitor ( sort keys( %{ $monitors{DIRECTOR}{$director} } ) ) {
        my $file            = "";
        my $start_timestamp = "";
        my $start_time      = "";
        my $last_timestamp  = "";
        my $last_time       = "";

        if ( defined $monitors{'DIRECTOR'}{$director}{$monitor}{'FILE'} && $monitors{'DIRECTOR'}{$director}{$monitor}{'FILE'} ne '' ) {
          $file = $monitors{'DIRECTOR'}{$director}{$monitor}{'FILE'};
        }
        if ( defined $monitors{'DIRECTOR'}{$director}{$monitor}{'START_TIMESTAMP'} && isdigit( $monitors{'DIRECTOR'}{$director}{$monitor}{'START_TIMESTAMP'} ) ) {
          $start_timestamp = $monitors{'DIRECTOR'}{$director}{$monitor}{'START_TIMESTAMP'};
        }
        if ( defined $monitors{'DIRECTOR'}{$director}{$monitor}{'START_TIME'} && $monitors{'DIRECTOR'}{$director}{$monitor}{'START_TIME'} ne '' ) {
          $start_time = $monitors{'DIRECTOR'}{$director}{$monitor}{'START_TIME'};
        }
        if ( defined $monitors{'DIRECTOR'}{$director}{$monitor}{'LAST_TIMESTAMP'} && isdigit( $monitors{'DIRECTOR'}{$director}{$monitor}{'LAST_TIMESTAMP'} ) ) {
          $last_timestamp = $monitors{'DIRECTOR'}{$director}{$monitor}{'LAST_TIMESTAMP'};
        }
        if ( defined $monitors{'DIRECTOR'}{$director}{$monitor}{'LAST_TIME'} && $monitors{'DIRECTOR'}{$director}{$monitor}{'LAST_TIME'} ne '' ) {
          $last_time = $monitors{'DIRECTOR'}{$director}{$monitor}{'LAST_TIME'};
        }

        if ( defined $file && defined $start_timestamp && defined $start_time && defined $last_timestamp && defined $last_time ) {
          if ( $monitor =~ m/VIRTUAL_VOLUMES_PERPETUAL_MONITOR\.log$/ ) {
            parse_vvol_monitor( "$director", "$file", "$start_timestamp", "$start_time", "$last_timestamp", "$last_time" );
          }
          if ( $monitor =~ m/PERPETUAL_vplex_sys_perf_mon\.log$/ ) {
            parse_sys_perf_monitor( "$director", "$file", "$start_timestamp", "$start_time", "$last_timestamp", "$last_time" );
          }
        } ## end if ( defined $file && ...)
      } ## end foreach my $monitor ( sort ...)
    } ## end foreach my $director ( sort...)
  } ## end if ( defined $monitors...)

  return 1;
} ## end sub parse_monitors

sub parse_sys_perf_monitor {
  my $director        = shift;
  my $file            = shift;
  my $start_timestamp = shift;
  my $start_time      = shift;
  my $last_timestamp  = shift;
  my $last_time       = shift;

  if ( -f "$file" ) {
    open( MON, "< $file" ) || error( "Couldn't open file $file $!" . __FILE__ . ":" . __LINE__ ) && exit;

    my $idx = 0;
    my @header;
    my $perf_timestamp = "";
    my $perf_date      = "";
    my $perf_time      = "";

    while ( my $line = <MON> ) {
      chomp $line;
      $idx++;

      # header
      if ( $idx == 1 ) {
        @header = split( ",", $line );
        next;
      }

      # stat time
      my ( $stat_time, undef ) = split( ",", $line );    # 2017-07-14 06:12:43,1500012763372,0,0,39525,42924,...

      ( $perf_date, $perf_time ) = split( /\s+/, $stat_time );
      my ( $year, $mon, $mday ) = split( /-/, $perf_date );
      my ( $hour, $min, $sec )  = split( /:/, $perf_time );

      $perf_timestamp = timelocal( $sec, $min, $hour, $mday, $mon - 1, $year );

      if ( defined $perf_timestamp && isdigit($perf_timestamp) && $perf_timestamp > $start_timestamp ) {

        #print "$file : $perf_date : $perf_time\n";

        my $value_idx = 0;
        my @values    = split( ",", $line );
        my $vvol_name = "";
        foreach my $value (@values) {
          chomp $value;

          if ( $value_idx == 0 ) {
            $vvol_name = $value;
          }
          $data{'perf'}{$director}{'director'}{$perf_timestamp}{ $header[$value_idx] } = $value;
          $value_idx++;
        } ## end foreach my $value (@values)
        $data{'perf'}{$director}{'director'}{$perf_timestamp}{'STAT_TIME'}      = "$perf_timestamp : $perf_date $perf_time";
        $data{'perf'}{$director}{'director'}{$perf_timestamp}{'LAST_TIME'}      = $last_time;
        $data{'perf'}{$director}{'director'}{$perf_timestamp}{'LAST_TIMESTAMP'} = $last_timestamp;
      } ## end if ( defined $perf_timestamp...)
    } ## end while ( my $line = <MON> )

    close(MON);
  } ## end if ( -f "$file" )

  return 1;
} ## end sub parse_sys_perf_monitor

sub parse_vvol_monitor {
  my $director        = shift;
  my $file            = shift;
  my $start_timestamp = shift;
  my $start_time      = shift;
  my $last_timestamp  = shift;
  my $last_time       = shift;

  if ( -f "$file" ) {
    open( MON, "< $file" ) || error( "Couldn't open file $file $!" . __FILE__ . ":" . __LINE__ ) && exit;

    my $idx = 0;
    my @header;
    my $perf_timestamp = "";
    my $perf_date      = "";
    my $perf_time      = "";

    while ( my $line = <MON> ) {
      chomp $line;
      $idx++;

      # header
      if ( $idx == 1 ) {
        @header = split( ",", $line );
        next;
      }

      # stat time
      if ( $line =~ m/^2[0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9] [0-9][0-9]:[0-9][0-9]:[0-9][0-9]$/ ) {    # statistic time (e.g. 2017-07-12 00:09:08)
        ( $perf_date, $perf_time ) = split( /\s+/, $line );
        my ( $year, $mon, $mday ) = split( /-/, $perf_date );
        my ( $hour, $min, $sec )  = split( /:/, $perf_time );

        $perf_timestamp = timelocal( $sec, $min, $hour, $mday, $mon - 1, $year );
        next;
      } ## end if ( $line =~ m/^2[0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9] [0-9][0-9]:[0-9][0-9]:[0-9][0-9]$/)

      if ( defined $perf_timestamp && isdigit($perf_timestamp) && $perf_timestamp > $start_timestamp ) {

        #print "$file : $perf_date : $perf_time\n";

        my $value_idx = 0;
        my @values    = split( ",", $line );
        my $vvol_name = "";
        foreach my $value (@values) {
          chomp $value;

          if ( $value_idx == 0 ) {
            $vvol_name = $value;
          }
          $data{'perf'}{'virtual volumes'}{$vvol_name}{$director}{$perf_timestamp}{ $header[$value_idx] } = $value;
          $value_idx++;
        } ## end foreach my $value (@values)
        $data{'perf'}{'virtual volumes'}{$vvol_name}{$director}{$perf_timestamp}{'STAT_TIME'}      = "$perf_timestamp : $perf_date $perf_time";
        $data{'perf'}{'virtual volumes'}{$vvol_name}{$director}{$perf_timestamp}{'LAST_TIME'}      = $last_time;
        $data{'perf'}{'virtual volumes'}{$vvol_name}{$director}{$perf_timestamp}{'LAST_TIMESTAMP'} = $last_timestamp;
      } ## end if ( defined $perf_timestamp...)
    } ## end while ( my $line = <MON> )

    close(MON);
  } ## end if ( -f "$file" )

  return 1;
} ## end sub parse_vvol_monitor

sub download_monitors {
  my @monitors = `$SSH $vplex_ssh_user\@$vplex_ssh_ip -p $vplex_ssh_port "LC_ALL=en_US LANG=en_US LC_NUMERIC=en_US ls -l /var/log/VPlex/cli/director-*_PERPETUAL_vplex_sys_perf_mon.log /var/log/VPlex/cli/director-*_VIRTUAL_VOLUMES_PERPETUAL_MONITOR.log" 2>\&1`;

  foreach my $line (@monitors) {
    chomp $line;

    #-rw-r--r-- 1 service users 7125568 Jul 12 08:03 /var/log/VPlex/cli/director-1-1-A_PERPETUAL_vplex_sys_perf_mon.log

    my ( undef, undef, undef, undef, $size, undef, undef, undef, $file ) = split( /\s+/, $line );
    my $filename = $file;
    $filename =~ s/^\/var\/log\/VPlex\/cli\///;

    #print "$file\n";

    my $scp_cmd = "scp -P $vplex_ssh_port $vplex_ssh_user\@$vplex_ssh_ip:$file $monitors_dir";

    #print "$scp_cmd\n";

    my $scp_out = `$scp_cmd`;

    #print "$scp_out\n";

    if ( -f "$monitors_dir/$filename" ) {
      my $director = $filename;
      $director =~ s/_VIRTUAL_VOLUMES_PERPETUAL_MONITOR\.log//;
      $director =~ s/_PERPETUAL_vplex_sys_perf_mon\.log//;

      my $file_size = ( stat("$monitors_dir/$filename") )[9];
      if ( isdigit($file_size) && $file_size == 0 ) {
        error("WARNING : This performance data source is empty! \"$filename\"");
        error("WARNING : Problem description and fix: http://www.stor2rrd.com/VPLEX-monitor-stopped.htm");
      }

      my $file_last_stats_time = "";
      if ( $filename =~ m/_VIRTUAL_VOLUMES_PERPETUAL_MONITOR.log$/ ) {
        $file_last_stats_time = "";
        $file_last_stats_time = `grep '^2[0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9] [0-9][0-9]:[0-9][0-9]:[0-9][0-9]\$' "$monitors_dir/$filename" | tail -1`;
        chomp $file_last_stats_time;
      }
      if ( $filename =~ m/_PERPETUAL_vplex_sys_perf_mon.log$/ ) {
        my $file_last_line = `tail -1 "$monitors_dir/$filename"`;
        chomp $file_last_line;
        $file_last_stats_time = "";
        ( $file_last_stats_time, undef ) = split( ",", $file_last_line );
      } ## end if ( $filename =~ m/_PERPETUAL_vplex_sys_perf_mon.log$/)

      if ( defined $file_last_stats_time && $file_last_stats_time ne '' ) {
        my ( $stat_date, $stat_time ) = split( /\s+/, $file_last_stats_time );
        my ( $year, $mon, $mday ) = split( /-/, $stat_date );
        my ( $hour, $min, $sec )  = split( /:/, $stat_time );
        my $stat_timestamp = timelocal( $sec, $min, $hour, $mday, $mon - 1, $year );

        # start time to collect data
        my $start_timestamp = $stat_timestamp - $sample_rate;
        my $start_time      = localtime($start_timestamp);

        $monitors{'DIRECTOR'}{$director}{$filename}{'FILE'}            = "$monitors_dir/$filename";
        $monitors{'DIRECTOR'}{$director}{$filename}{'START_TIMESTAMP'} = "$start_timestamp";
        $monitors{'DIRECTOR'}{$director}{$filename}{'START_TIME'}      = "$start_time";
        $monitors{'DIRECTOR'}{$director}{$filename}{'LAST_TIMESTAMP'}  = "$stat_timestamp";
        $monitors{'DIRECTOR'}{$director}{$filename}{'LAST_TIME'}       = "$file_last_stats_time";

        #print "$filename : $file_last_stats_time : $stat_timestamp\n";
        #print "start : $start_timestamp : $start_time\n";
        #print "\n";
      } ## end if ( defined $file_last_stats_time...)
    } ## end if ( -f "$monitors_dir/$filename")
  } ## end foreach my $line (@monitors)

  return 1;
} ## end sub download_monitors

sub new_post_request_configdump {
  my $url       = shift;
  my $json_body = shift;

  my $agent = LWP::UserAgent->new( ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 } );
  my $request = HTTP::Request->new( POST => $url );

  $request->header(
    'Content-Type' => 'application/json',
    'username'     => "$user",
    'password'     => "$pass"
  );

  $request->content($json_body);

  my $results = $agent->request($request);

  if ( !$results->is_success ) {
    my $st_line = $results->status_line;
    error("Request error: $st_line");
    error("POST url     : $url");
    error("json body    : $json_body");

    return 0;
  } ## end if ( !$results->is_success)

  my $result = decode_json( $results->content );

  # workaround for downloading configdump file
  if ( defined $result->{'response'}{'message'} && $result->{'response'}{'message'} =~ m/Command execution taking longer than 60 seconds/ ) {
    if ( defined $results->header('location') && $results->header('location') ne "" ) {
      my $res_loc = $results->header('location');
      return "Command execution taking longer than 60 seconds,$res_loc";
    }
  } ## end if ( defined $result->...)

  #print Dumper $results->header('location');

  return $result;
} ## end sub new_post_request_configdump

sub new_get_request_configdump {
  my $url = shift;

  my $agent = LWP::UserAgent->new( ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 } );
  my $request = HTTP::Request->new( GET => $url );

  $request->header(
    'Content-Type' => 'application/json',
    'username'     => "$user",
    'password'     => "$pass"
  );

  my $results = $agent->request($request);

  if ( !$results->is_success ) {
    my $st_line = $results->status_line;

    # workaround for downloading configdump file
    if ( $st_line =~ m/517 Unknown code/ && $url =~ m/\/vplex\/commandresult\// ) {    # file is not complete now
      my $result = decode_json( $results->content );
      if ( defined $result->{'response'}{'message'} && $result->{'response'}{'message'} =~ m/Command execution continues to be under progress/ ) {
        return "Command execution continues to be under progress";
      }
    } ## end if ( $st_line =~ m/517 Unknown code/...)

    error("Request error: $st_line");
    error("GET url      : $url");

    return 0;
  } ## end if ( !$results->is_success)

  my $result = decode_json( $results->content );

  return $result;
} ## end sub new_get_request_configdump

sub new_post_request {
  my $url       = shift;
  my $json_body = shift;

  my $agent = LWP::UserAgent->new( ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 } );
  my $request = HTTP::Request->new( POST => $url );

  $request->header(
    'Content-Type' => 'application/json',
    'username'     => "$user",
    'password'     => "$pass"
  );

  $request->content($json_body);

  my $results = $agent->request($request);

  if ( !$results->is_success ) {
    my $st_line = $results->status_line;
    error("Request error: $st_line");
    error("POST url     : $url");
    error("json body    : $json_body");

    return 0;
  } ## end if ( !$results->is_success)

  my $result = decode_json( $results->content );

  return $result;
} ## end sub new_post_request

sub new_get_request {
  my $url = shift;

  my $agent = LWP::UserAgent->new( ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 } );
  my $request = HTTP::Request->new( GET => $url );

  $request->header(
    'Content-Type' => 'application/json',
    'username'     => "$user",
    'password'     => "$pass"
  );

  my $results = $agent->request($request);

  if ( !$results->is_success ) {
    my $st_line = $results->status_line;
    error("Request error: $st_line");
    error("GET url      : $url");

    return 0;
  } ## end if ( !$results->is_success)

  my $result = decode_json( $results->content );

  return $result;
} ## end sub new_get_request

sub isdigit {
  my $digit = shift;

  if ( !defined($digit) || $digit eq '' ) {
    return 0;
  }

  my $digit_work = $digit;
  $digit_work =~ s/[0-9]//g;
  $digit_work =~ s/\.//;
  $digit_work =~ s/^-//;
  $digit_work =~ s/e//;
  $digit_work =~ s/\+//;
  $digit_work =~ s/\-//;

  if ( length($digit_work) == 0 ) {

    # is a number
    return 1;
  }

  # NOT a number
  return 0;
} ## end sub isdigit

sub file_time_diff {
  my $file = shift;

  my $act_time  = time();
  my $file_time = $act_time;
  my $time_diff = 0;

  if ( -f $file ) {
    $file_time = ( stat($file) )[9];
    $time_diff = $act_time - $file_time;
  }

  return ($time_diff);
} ## end sub file_time_diff

sub print_info {
  my $text     = shift;
  my $act_time = localtime();
  chomp($text);

  print "$act_time: $text\n";

  return 1;
} ## end sub print_info

# error handling
sub error {
  my $text     = shift;
  my $act_time = localtime();
  chomp($text);

  #print "ERROR          : $text : $!\n";
  print STDERR "$act_time: $text : $!\n";

  #print "$act_time: $text : $!\n" if $DEBUG > 2;;

  return 1;
} ## end sub error
